Program PDICReader;
{$PMTYPE PM}
{$R-,S-,I+}
uses OS2DEF, OS2BASE, OS2PMAPI, DOS, strings, VPUtils;
{$R PMLEXS21}

const
  HeaderLen     = 256;
  HeaderNameLen = 100;
  LexTitleLen   = 40;
  MaxIndex      = 4097;
  IndexBufferLen= 40960;
  BlockBufferLen= 8192;
  MaxDispItemNum= 20;
  MaxDispItemNum1= 21;
  MaxLex        = 8;

  ClassName     = 'PMLEXSRCH';
  ID_Timer      = 1;
  RetryInterval = 200;
  MaxPending    = 10;

  ControlProgramID = 100;
  IDM_File      = 200;
  IDM_SearchTrans = 201;
  IDM_Pref      = 300;
  IDM_RunMode   = 301;
  IDM_RunSetup  = 302;
  IDM_Edit      = 400;
  IDM_AddWord   = 401;
  IDM_DelWord   = 402;
  IDM_EditWord  = 403;
  IDM_About     = 500;
  IDM_SearchClipBoardText = 600;
  IDD_SearchWord= 101;
  IDD_Result    = 102;
  IDD_SearchTransDlg = 103;
  IDD_SearchTransEntry = 104;
  IDD_SearchTransResultDlg = 105;
  IDD_SearchTransResult = 106;
  IDD_RunMode   = 200;
  IDD_SetIncremental = 201;
  IDD_Edit      = 300;
  IDD_EditTitle = 301;
  IDD_EditWord  = 302;
  IDD_EditResult= 303;
  IDD_EditDetail= 304;
  IDD_DetailDlg = 400;
  IDD_Detail    = 401;
  IDD_About     = 500;

  MLEBufSize    = 4096;
  TransMLEBufSize = 40960;

  WM_USER_Found = WM_USER;
  WM_USER_TransFound = WM_USER + 1;
  version       = '2.1';
  ConfigName    = 'PMLEXS.CFG';
  Configurator  = 'LEXCNF11.EXE ';

type
  LexTypes      = (tinyLex, smallLex, mediumLex, largeLex);
  WordOrder     = (ByCode, IgnoreCase, ByDictionary, Dessending);
  LexHeader     = record
    HeaderName  : array[1..HeaderNameLen] of char;
    LexTitle    : array[1..LexTitleLen] of char;
    version     : word;
    MaxWordLen  : word;
    MaxTransLen : word;
    BlockSize   : word;
    IndexBlockNum : word;
    HeaderSize  : word;
    IndexSize   : word;
    EmptyBlock  : integer;
    nIndex      : word;
    nBlock      : word;
    nWord       : word;
    LexOrder    : WordOrder;
    LexType     : LexTypes;
    attrLen     : byte;
    dummy       : string[88]
  end;

  IndexBuffer   = array[1..IndexBufferLen] of byte;

  pstring       = ^string;
  IndexTag      = record
    FirstWord   : pstring;      (* First word in the Block *)
    BlockNum    : integer       (* Record Number of the Block *)
  end;

  IndexTagAry   = array[0..MaxIndex] of IndexTag;

  SearchStat    = record        (* Searching State *)
    currentStr,                 (* string start at currentPos *)
    prevstr     : string;       (* string preceeding currentStr *)
    index       : word;         (* index *)
    num         : word;         (* block number *)
    flen        : word;         (* current field length *)
    currentPos  : word;         (* current Pointer *)
    extPos      : word;         (* extended data pointer *)
    extAtr      : byte          (* extended data attribute *)
  end;

  BlockBuffer   = array[1..BlockBufferLen] of byte;

  FoundItem     = record
    Eng, Jap, Example : string;
    extBlock    : word;         (* BlockNum of extended data *)
    extPos      : word;         (* extended data pointer *)
    extAtr      : byte;         (* extended data attribute *)
    LexNum      : byte;         (* Number of Lexikon *)
    endpos      : word          (* end position in MLE *)
  end;

  pTransFoundList   = ^TransFoundList;
  TransFoundList    = record
    Eng, Trans  : string;
    next        : pTransFoundList
  end;

  SearchBuffer  = object
    lexname     : string;       (* Name of lexicon file *)
    lexfile     : file;         (* Lexicon file *)
    header      : LexHeader;    (* Header of Lexicon *)
    Indeces     : IndexTagAry;  (* Index of Lexicon *)
    buf         : BlockBuffer;  (* buffer for data block *)
    sStat       : SearchStat;   (* current status *)
    TransStr    : string;       (* Translation *)
    ToReadTransStr : boolean;   (* Read Translation? *)
    constructor Init(initLexName : string);    (* Initialisation *)
    procedure   OpenLex(initLexName : string); (* Open lexicon file *)
    procedure   LoadHeader;     (* Load lexicon header *)
    procedure   LoadIndeces; virtual;          (* Load lexicon index *)
    procedure   LoadBlock(BlockToLoad : word);
    procedure   getBasicItems;  (* get an item at sStat.currentPos *)
    procedure   SearchFirst(WordToSearch : string;
                            var ToSearch : SearchStat);
      (* Locate ToSearch and sStat at the word not less than WordToSearch *)
    procedure   GetData(var SearchResult : FoundItem;
                        var avail : boolean);
      (* Get Word at sStat.currentPos *)
    procedure   GetExample(var SearchResult : FoundItem);
      (* Get Expanded item - Example *)
    procedure   SearchTranslate(TranslateToSearch : string); virtual;
    procedure   ResetStat;
    destructor  Done; virtual;
  end;

  pSearchBuffer = ^SearchBuffer;

  BlockSet      = array [0..15] of set of 0..255;

  EditableSearchBuffer = object(SearchBuffer)
    UsedBlock   : BlockSet;   (* map of block used *)
    SemHandle   : HMTX;
    constructor Init(initLexName : string);
    procedure   LoadIndeces; virtual;
    procedure   SearchTranslate(TranslateToSearch : string); virtual;
    procedure   ChangeIdxWord(NewIdxWord : string;
                              IndexToChange : integer);
    function    GetActualBlockSize : integer;
    procedure   AddIndex(IdxWordToAdd : string;
                         var BlockToAdd: word;
                         var NewIndex  : integer);
    procedure   AddToBlock(WordToAdd,
                           TransToAdd,
                           ExampleToAdd : string; var avail : boolean);
    procedure   DivideBlock(var avail : boolean);
    procedure   AddAWord(WordToAdd,
                         TransToAdd,
                         ExampleToAdd : string);
    procedure   SetSearchStatToTailOfBlock;
    procedure   DelIndex(IndexToDel  : integer);
    procedure   DelFromBlock(var NewNextWord : string);
    procedure   DelAWord(WordToDel : string);
    procedure   SaveHeader;
    procedure   SaveIndex;
    procedure   SaveBlock(NewBlockNum : word);
  end;

  pEditableSearchBuffer = ^EditableSearchBuffer;


  LexIndex      = 1..MaxLex;
  AdditionalLex = 2..MaxLex;
  AdditionalLex0= 0..MaxLex;
  SearchStatAry = array[LexIndex] of SearchStat;
  DispItemNums  = 0..MaxDispItemNum1;
  RunModes      = (NormalMode, IncMode, ChildMode);
  LexNameAry    = array[LexIndex] of string;
  settings      = record
    DispItemNum : DispItemNums;
    RunMode     : RunModes;
    lexname     : LexNameAry;               (* Names of Lexicon files *)
    lexnum      : integer;                  (* Number of Lexicon *)
    homedir     : DirStr;
    x0, y0,
    x1, y1      : word
  end;

  TransMLEBuffer = array[1..TransMLEBufSize] of char;
  pTrans      = ^TransMLEBuffer;


var
  setting     : settings;
  Lex1        : SearchBuffer;               (* Primary Lexicon *)
  LexAdd      : array[AdditionalLex] of pSearchBuffer;  (* Additional Lexicon *)
  LexToAdd    : pEditableSearchBuffer;      (* Lexicon accepting record/delete *)
  LexToAddNum : AdditionalLex0;             (* Number of Editable Lexicon *)
  SearchHist  : SearchStatAry;              (* Current Search Status *)
  NullStr,
  MaxStr      : pstring;                    (* Pointers to special strings *)
  OldEng,
  wd,
  prewd,                                    (* Current/Previous word to search *)
  twd         : string;
  resultNum   : DispItemNums;               (* Number of results found *)
  results     : array[DispItemNums] of FoundItem;    (* Results found *)
  ThreadID,
  TransThreadID : TID;
  FoundTrans  : pTransFoundList;

(* PM *)
  myhab       : HAB;
  myhmq       : HMQ;
  hwndFrame,
  hwndClient,
  hwndMenu    : HWND;
  myqmsg      : QMSG;
  sw          : SWCNTRL;
  MLEBuffer   : array[1..MLEBufSize] of char;
  MLEDataPos  : IPT;
  MLEBufferTrans : pTrans;
  TimerID     : ULONG;
  SearchRunning,
  TransSearchRunning,
  CurrentFocusOnEntry : boolean;
  Pending     : byte;

function LoUShort(w : ULONG) : UShort; inline;
begin
  LoUShort := SmallWord( w )
end;

function HiUShort(w : ULONG) : UShort; inline;
begin
  HiUShort := SmallWord( w shr 16 )
end;

procedure ResetSearchStat(var s : SearchStat);
begin
  with s do begin
    flen  := 0;
    index := 0;
    num   := maxint;
    currentPos := 1;
    currentStr := MaxStr^;
    prevStr := ''
  end
end;

constructor SearchBuffer.Init(initLexName : string);
begin
  system.filemode := 0;
  OpenLex(initLexName)
end;

procedure SearchBuffer.OpenLex(initLexName : string);
begin
  lexname := initLexName;
  assign(lexfile, lexname);
  reset(lexfile, 1);
  fillchar(Indeces, sizeof(Indeces), chr(0));
  LoadHeader;
  LoadIndeces;
  ResetStat
end;

procedure SearchBuffer.LoadHeader;
begin
  blockread(lexfile, header, sizeof(header));
  with header do begin
    case version  div $100 of
      0  : begin                                (* old *)
          nWord        := nBlock;
          IndexSize    := headersize;
          headersize   := 256;
          IndexBlockNum:= BlockSize;
          BlockSize    := MaxTransLen
        end;
      1  :;                                     (* unknown *)
      2  :;                                     (* NEWDIC *)
      3  : BlockSize := BlockSize * 16          (* NEWDIC2 *)
    end
  end
end;

procedure SearchBuffer.LoadIndeces;
var
  idx, n : word;
  i, ii,
  strlen : word;
  idxbuf : ^IndexBuffer;
begin
  new(idxbuf);
  seek(lexfile, HeaderLen);
  blockread(lexfile, idxbuf^, header.IndexSize);
  with Indeces[0] do begin
    BlockNum := 0;
    FirstWord := NullStr
  end;
  idx := 1;
  n := idxbuf^[1] + idxbuf^[2] * 256;
  i := 3;
  while (n <> 0) or (idxbuf^[i] <> 0) do begin
    ii  := i;
    while idxbuf^[ii]<>0 do inc(ii);
    inc(ii);
    strlen := ii - i;
    if strlen > 256 then strlen := 256;
    with Indeces[idx] do begin
      BlockNum := n;
      getmem(FirstWord, strlen);
      move(idxbuf^[i], FirstWord^[1], pred(strlen));
      FirstWord^[0] := chr(pred(strlen))
    end;
    inc(idx);
    i := ii;
    n := idxbuf^[i] + idxbuf^[succ(i)] * 256;
    inc(i, 2)
  end;
  dispose(idxbuf);
  with Indeces[idx] do begin
    BlockNum := idx;
    FirstWord:= MaxStr
  end
end;

procedure SearchBuffer.getBasicItems;
const ExtendedFormat = $10;
var current, j, k : word;
    strlen : word;
    atr, cmp  : byte;
begin
  with sStat do begin
    flen    := buf[currentPos] + buf[succ(currentPos)] * 256;
    if flen = 0 then begin
      currentStr := MaxStr^;
      exit
    end;
    current := currentPos;
    cmp     := buf[current + 2];
    inc(current, 3);
    j       := current;
    while buf[j]<>0 do inc(j);
    strlen  := j - current;
    move(buf[current], currentstr[1], strlen);
    currentstr[0] := chr(strlen);
    currentstr    := copy(prevstr, 1, cmp) + currentstr;
    if ToReadTransStr then begin
      atr   := buf[succ(j)];
      inc(j, 2);
      if atr and ExtendedFormat = 0 then begin
        strlen := flen - strlen - 2;
            (* one for atr, one for terminating zero char *)
        extAtr := 0
      end else begin
        k   := j;
        while buf[k]<>0 do inc(k);
        strlen := k - j;
        extPos := succ(k);
        extAtr := buf[extPos]
      end;
      if strlen > 255 then strlen := 255;
      move(buf[j], transstr[1], strlen);
      transstr[0] := chr(strlen)
    end
  end
end;

procedure SearchBuffer.LoadBlock(BlockToLoad : word);
begin
  with header do begin
{$I-}
    seek(lexfile, HeaderLen + IndexSize + BlockSize * BlockToLoad);
    blockread(lexfile, buf, BlockSize)
{$I+}
  end;
  if ioresult <> 0 then fillchar(buf, sizeof(buf), 0)
end;

procedure SearchBuffer.SearchFirst(WordToSearch : string;
                                   var ToSearch : SearchStat);
var n       : word;
    CurrentRec : word;

  function SetIndex(root : word) : word;
   (* binary search. non-recursive *)
  var floor, ceil,
      inum : word;
      state: (ok, tooSmall, tooLarge);
  begin
    if root = 0 then root := 1;
    floor := 0;
    ceil  := succ(header.nIndex);
    if      WordToSearch < Indeces[root].FirstWord^ then
      state := tooLarge
    else if WordToSearch >=Indeces[succ(root)].FirstWord^ then
      if    WordToSearch < Indeces[root + 2].FirstWord^ then begin
        inc(root);
        state := ok
      end else
        state := tooSmall
    else
      state := ok;
    while state <> ok do begin
      if state = tooSmall then floor := root
                          else ceil  := root;
      root := (floor + ceil) shr 1;
      if      WordToSearch < Indeces[root].FirstWord^ then
        state := tooLarge
      else if WordToSearch >=Indeces[succ(root)].FirstWord^ then
        state := tooSmall
      else
        state := ok
    end;
    if root = 0 then root := 1;
    SetIndex := root
  end; { SearchFirst.SetIndex }

begin  { SearchFirst }
  CurrentRec := sStat.num;
  if (ToSearch.index <> 0) and
     (WordToSearch > ToSearch.currentStr) then sStat := ToSearch;
  with sStat do begin
    if WordToSearch < currentStr then begin
      index   := 0;
      prevstr := '';
      currentPos := 1
    end;
    index := SetIndex(index);
    if CurrentRec <> Indeces[index].BlockNum then begin
      CurrentRec := Indeces[index].BlockNum;
      LoadBlock(CurrentRec)
    end;
    if CurrentRec <> num then begin
      num := CurrentRec;
      prevstr := '';
      currentPos := 1
    end;
    ToReadTransStr := false;
    GetBasicItems;
    while (currentStr < WordToSearch) and (flen <> 0) do begin
      inc(currentPos, flen + 3);
      prevstr := currentStr;
      GetBasicItems
    end
  end;
  ToSearch := sStat
end;  { SearchFirst }

procedure SearchBuffer.GetData(var SearchResult : FoundItem;
                               var avail : boolean);

  procedure LoadNextBlock;
  begin
    with sStat do begin
      inc(index);
      num   := Indeces[index].BlockNum;
      LoadBlock(num);
      prevStr    := '';
      currentPos := 1;
      GetBasicItems
    end
  end;  { GetData.LoadNextBlock }

begin  { GetData }
  ToReadTransStr := true;
  GetBasicItems;
  with sStat do begin
    if (flen = 0) and
       (index < header.nIndex) then LoadNextBlock;
    if flen <> 0 then begin
      with SearchResult do begin
        eng := currentStr;
        jap := transStr;
        extBlock := num;
        extPos   := sStat.extPos;
        extAtr   := sStat.extAtr
      end;
      inc(currentPos, flen + 3);
      prevstr := currentstr;
      avail := true
    end else
      avail := false
  end
end;  { GetData }

procedure SearchBuffer.GetExample(var SearchResult : FoundItem);
var reserve   : BlockBuffer;
    startPos,
    strlen, i : word;
begin
  with SearchResult do begin
    if extBlock <> sStat.num then begin
      reserve := buf;
      LoadBlock(extBlock)
    end;
    if (extAtr and $10) = 0 then begin
      startPos := succ(extPos);
      i := startPos;
      while buf[i]<>0 do inc(i);
      strlen  := i - startPos;
      if strlen > 255 then strlen := 255;
      move(buf[startPos], Example[1], strlen);
      Example[0] := chr(strlen)
    end;
    if extBlock <> sStat.num then buf := reserve
  end
end;

procedure SearchBuffer.SearchTranslate(TranslateToSearch : string);
var buf2      : BlockBuffer;  (* buffer for data block *)
    sStat2    : SearchStat;
    TransStr2 : string;
    p         : pTransFoundList;

  procedure LoadBlock2(BlockToLoad : longint);
  begin
    with header do begin
      seek(lexfile, HeaderLen + IndexSize + BlockSize * BlockToLoad);
      blockread(lexfile, buf2, BlockSize)
    end
  end;

  procedure getBasicItems2;
  const ExtendedFormat = $10;
  var current, j, k : word;
      strlen : word;
      atr, cmp  : byte;
  begin
    with sStat2 do begin
      flen    := buf2[currentPos] + buf2[succ(currentPos)] * 256;
      if flen = 0 then begin
        currentStr := MaxStr^;
        exit
      end;
      current := currentPos;
      cmp     := buf2[current + 2];
      inc(current, 3);
      j       := current;
      while buf2[j]<>0 do inc(j);
      strlen  := j - current;
      move(buf2[current], currentstr[1], strlen);
      currentstr[0] := chr(strlen);
      currentstr    := copy(prevstr, 1, cmp) + currentstr;
      atr   := buf2[succ(j)];
      inc(j, 2);
      if atr and ExtendedFormat = 0 then begin
        strlen := flen - strlen - 2;
            (* one for atr, one for terminating zero char *)
        extAtr := 0
      end else begin
        k   := j;
        while buf2[k]<>0 do inc(k);
        strlen := k - j
      end;
      if strlen > 255 then strlen := 255;
      move(buf2[j], TransStr2[1], strlen);
      TransStr2[0] := chr(strlen)
    end
  end;

begin
  with sStat2 do
    for index := 1 to header.nIndex do begin
      DosSleep(1);
      LoadBlock2(Indeces[index].BlockNum);
      prevstr := '';
      currentPos := 1;
      GetBasicItems2;
      while flen <> 0 do begin
        if pos(TranslateToSearch, TransStr2)<>0 then begin
          new(p);
          with p^ do begin
            Eng := CurrentStr;
            Trans := TransStr2;
            next := FoundTrans
          end;
          FoundTrans := p
        end;
        inc(currentPos, flen + 3);
        prevstr := currentStr;
        GetBasicItems2
      end
    end
end;

procedure SearchBuffer.ResetStat;
begin
  ResetSearchStat(sStat)
end;

destructor SearchBuffer.Done;
var i, j : integer;
begin
  for i:= 1 to header.nIndex do
    with Indeces[i] do begin
      j := length(FirstWord^);
      freemem(FirstWord, succ(j))
    end
end;

(* Set of Used Block operations *)
procedure AddBlockToSet(var s : BlockSet; n : word);
begin
  s[hi(n)] := s[hi(n)] + [lo(n)]
end;

procedure DelBlockFromSet(var s : BlockSet; n : word);
begin
  s[hi(n)] := s[hi(n)] - [lo(n)]
end;

function  FindFirstEmptyBlock(var s : BlockSet) : integer;
var i : word;
begin
  i := 0;
  while lo(i) in s[hi(i)] do inc(i);
  FindFirstEmptyBlock := i
end;

(* String Manipuration *)
function CompareStr(prevs, news : string) : integer;
(* Compare prevs and news and returns first position
of charactor which different between prevs and news *)
var i, j : integer;
begin
  if length(prevs) < length(news) then i := length(prevs)
                                  else i := length(news);
  j := 1;
  while (prevs[j] = news[j]) and (j <= i) do inc(j);
  CompareStr := pred(j)
end;

constructor EditableSearchBuffer.Init(initLexName : string);
begin
  system.filemode := 2;
  OpenLex(initLexName);
  DosCreateMutexSem(nil, SemHandle, 0, false)
end;

procedure EditableSearchBuffer.LoadIndeces;
var i : word;
begin
  SearchBuffer.LoadIndeces;
  fillchar(UsedBlock, sizeof(UsedBlock), 0);
  for i := 1 to header.nIndex do
    AddBlockToSet(UsedBlock, Indeces[i].BlockNum)
end;

procedure EditableSearchBuffer.SearchTranslate(TranslateToSearch : string);
begin
  DosRequestMutexSem(SemHandle, 60000);
  SearchBuffer.SearchTranslate(TranslateToSearch);
  DosReleaseMutexSem(SemHandle)
end;

procedure EditableSearchBuffer.ChangeIdxWord(NewIdxWord : string;
                                             IndexToChange : integer);
(* change Index IndexToChange to NewIdxWord *)
begin
  with Indeces[IndexToChange] do begin
    freemem(FirstWord, succ(length(FirstWord^)));
    getmem(FirstWord, succ(length(NewIdxWord)));
    FirstWord^ := NewIdxWord
  end
end;

procedure EditableSearchBuffer.AddIndex(IdxWordToAdd : string;
                                        var BlockToAdd: word;
                                        var NewIndex  : integer);
(* Add New Index for new block.
  IdxWordToAdd : First word of new block,
  BlockToAdd   : Block number of new block,
  NewIndex     : Index number of new index,  < 0 : fail *)
var n : integer;
    i : word;
begin
  if header.nIndex >= MaxIndex then
    n := maxint
  else begin
    n := 0;
    for i := 1 to header.nIndex do
      inc(n, length(Indeces[i].FirstWord^) + 3)
      (* 2 : block number, 1 : null terminator *)
  end;
  if n > header.IndexSize - length(IdxWordToAdd) then begin
    NewIndex := -1;
    exit
  end;
  i := 0;
  repeat
    inc(i)
  until Indeces[i].FirstWord^ > IdxWordToAdd;
  move(Indeces[i], Indeces[succ(i)],
       (succ(header.nIndex) - pred(i)) * sizeof(IndexTag));
  BlockToAdd := FindFirstEmptyBlock(UsedBlock);
  AddBlockToSet(UsedBlock, BlockToAdd);
  with Indeces[i] do begin
    BlockNum := BlockToAdd;
    getmem(FirstWord, succ(length(IdxWordToAdd)));
    FirstWord^ := IdxWordToAdd
  end;
  NewIndex := i;
  inc(header.nIndex);
  inc(header.nBlock)
end;

function EditableSearchBuffer.GetActualBlockSize : integer;
(* Returns Current block size *)
var reserve : SearchStat;
begin
  reserve := sStat;
  with sStat do begin
    currentPos := 1;
    prevStr    := ''
  end;
  GetBasicItems;
  with sStat do begin
    while flen <> 0 do begin
      inc(currentPos, flen + 3);
      prevstr := currentstr;
      GetBasicItems
    end;
    GetActualBlockSize := pred(currentPos)
  end;
  sStat := reserve
end;

procedure EditableSearchBuffer.AddToBlock(WordToAdd,
                                          TransToAdd,
                                          ExampleToAdd :string;
                                          var avail : boolean);
(* Add new word at currentPos in current block
   WordToAdd : new word,
   TransToAdd : new translation,
   sStat supposed to point to insert new word *)
var
  eng     : string;
  ps      : word;
  actSize : word;
  len     : word;
  newflen : word;
  cmp,
  nextcmp,
  cmp2,
  cmpdiff : byte;
begin
  avail   := false;
  actSize := GetActualBlockSize;
  if actSize +
     length(WordToAdd) +
     length(TransToAdd) +
     length(ExampleToAdd) + 8 > header.BlockSize then exit;
  with sStat do begin
    if flen<>0 then begin
      nextcmp  := buf[currentPos + 2];
      cmp2     := CompareStr(WordToAdd, currentStr);
      cmpdiff  := cmp2 - nextcmp;
      if cmpdiff <> 0 then begin
        newflen:= flen - cmpdiff;
        move(newflen, buf[currentPos + cmpdiff], 2);
        buf[currentPos + 2 + cmpdiff] := cmp2
      end
    end else
      cmpdiff  := 0;
    cmp := CompareStr(prevstr, WordToAdd);
    eng := copy(WordToAdd, succ(cmp), length(WordToAdd) - cmp);
    len := 5 + length(eng) + length(TransToAdd);
          (* 2 : field length,
             1 : comp length,
             1 : attribute,
             1 : null char terminating english word *)
    if ExampleToAdd > '' then inc(len, length(ExampleToAdd) + 4);
          (* 1 : null char terminating translation,
             1 : attribute,
             1 : null char terminating example
             1 : null attribute *)
    move(buf[currentPos + cmpdiff], buf[currentPos + len],
         actSize - pred(currentPos) - cmpdiff);
    dec(len, 3);
    ps := currentPos;
    move(len, buf[ps], 2);
    inc(ps, 2);
    buf[ps] := cmp;
    inc(ps);
    move(eng[1], buf[ps], length(eng));
    inc(ps, length(eng));
    buf[ps] := 0;
    inc(ps);
    buf[ps] := $80;
    inc(ps);
    move(TransToAdd[1], buf[ps], length(TransToAdd));
    if ExampleToAdd > '' then  begin
      inc(buf[pred(ps)], $10);
      inc(ps, length(TransToAdd));
      buf[ps] := 0;
      inc(ps);
      buf[ps] := 1;
      inc(ps);
      move(ExampleToAdd[1], buf[ps], length(ExampleToAdd));
      inc(ps, length(ExampleToAdd));
      buf[ps] := 0;
      inc(ps);
      buf[ps] := $80
    end
  end;
  inc(header.nWord);
  avail := true
end;

procedure EditableSearchBuffer.SetSearchStatToTailOfBlock;
(* set insert position to tail of the current block *)
begin
  with sStat do begin
    currentPos := 1;
    prevStr    := ''
  end;
  GetBasicItems;
  with sStat do begin
    while flen <> 0 do begin
      inc(currentPos, flen + 3);
      prevstr := currentstr;
      GetBasicItems
    end
  end
end;

procedure EditableSearchBuffer.DivideBlock(var avail : boolean);
const
  LowerLimit = 0.4;
  UpperLimit = 0.6;
var newBuf  : BlockBuffer;
    lower,
    upper   : word;
    divideStr : string;
    dividePos : word;
    newBlock  : word;
    newIdx  : integer;
    newflen : word;
    cmp     : byte;
begin
  lower := trunc(header.BlockSize * LowerLimit);
  upper := round(header.BlockSize * UpperLimit);
  with sStat do begin
    currentPos := 1;
    prevStr    := ''
  end;
  GetBasicItems;
  with sStat do begin
    while (currentPos < lower) and (flen <> 0) do begin
      inc(currentPos, flen + 3);
      prevstr := currentstr;
      GetBasicItems
    end;
    divideStr := currentStr;
    dividePos := currentPos;
    while (currentPos < upper) and (flen <> 0) do begin
      if length(divideStr) > length(currentStr) then begin
        divideStr := currentStr;
        dividePos := currentPos
      end;
      inc(currentPos, flen + 3);
      prevstr := currentstr;
      GetBasicItems
    end
  end;
  AddIndex(divideStr, newBlock, newIdx);
  if newIdx < 0 then begin
    avail := false;
    exit
  end;
  newflen := buf[dividePos] + buf[succ(dividePos)] * 256;
  cmp     := buf[dividePos + 2];
  inc(newflen, cmp);
  fillchar(newBuf, sizeof(newBuf), 0);
  move(newflen, newBuf[1], 2);
  newBuf[3] := 0;
  move(divideStr[1], newBuf[4], cmp);
  move(buf[dividePos + 3], newBuf[cmp + 4],
                         header.BlockSize - pred(dividePos) - 3);
  with header do begin
    seek(lexfile, HeaderLen + IndexSize + BlockSize * newBlock);
    blockwrite(lexfile, newBuf, BlockSize)
  end;
  fillchar(buf[dividePos], sizeof(buf)-pred(dividePos), 0);
  SaveBlock(sStat.num);
  avail := true
end;

procedure EditableSearchBuffer.AddAWord(WordToAdd,
                                        TransToAdd,
                                        ExampleToAdd : string);
var newIndex : integer;
    newBlock : word;
    ToSearch : SearchStat;
    avail  : boolean;
begin
  DosRequestMutexSem(SemHandle, 60000);
  if header.nIndex = 0 then begin
    ResetStat;
    sStat.index := 1;
    AddIndex(WordToAdd, newBlock, newIndex);
    sStat.index := newIndex;
    sStat.num   := newBlock;
    fillchar(buf, sizeof(buf), 0)
  end else begin
    ResetSearchStat(ToSearch);
    with ToSearch do begin
      index := 0;
      SearchFirst(WordToAdd, ToSearch);
      if (flen = 0) and (index < header.nIndex) then
        with Indeces[succ(index)] do begin
          ResetSearchStat(ToSearch);
          index := 0;
          SearchFirst(FirstWord^, ToSearch)
        end;
      if (currentPos = 1) and (WordToAdd < currentStr) then
        if length(WordToAdd) < length(currentStr) then begin
          ChangeIdxWord(WordToAdd, index);
          ResetStat;
          sStat.index := index;
          sStat.num   := num;
          sStat.currentStr := currentStr
        end else begin
          if index = 1 then begin
            AddIndex(WordToAdd, newBlock, newIndex);
            if sStat.index < 0 then exit;
            ResetStat;
            sStat.index := newIndex;
            sStat.num := newBlock;
            fillchar(buf, sizeof(buf), 0)
          end else begin
            sStat.index := pred(index);
            sStat.num   := Indeces[sStat.index].BlockNum;
            LoadBlock(sStat.num);
            SetSearchStatToTailOfBlock
          end
        end
    end
  end;
  AddToBlock(WordToAdd, TransToAdd, ExampleToAdd, avail);
  if avail then
    SaveBlock(sStat.num)
  else begin
    DivideBlock(avail);
    if avail then AddAWord(WordToAdd, TransToAdd, ExampleToAdd)
  end;
  DosReleaseMutexSem(SemHandle)
end;

procedure EditableSearchBuffer.DelFromBlock(var NewNextWord : string);
(* delete current string *)
var
  nextStr : string;
  nextPos,
  newflen,
  nextflen: word;
  cmp,
  nextcmp,
  newcmp,
  cmpdiff : byte;
  i, j    : integer;
begin
  with sStat do begin
    nextPos  := currentPos + flen + 3;
    nextflen := buf[nextPos] + buf[succ(nextPos)] * 256;
    if nextflen <> 0 then begin
      nextcmp:= buf[nextPos + 2];
      j      := nextPos + 2;
      repeat inc(j) until buf[j] = 0;
      i      := j - nextPos - 3;
      move(buf[nextPos + 3], nextStr[1], i);
      nextStr[0] := chr(i);
      nextStr:= copy(currentStr, 1, nextcmp) + nextStr;
      newcmp := CompareStr(prevStr, nextStr);
      cmpdiff:= nextcmp - newcmp
    end else begin
      nextStr:= '';
      cmpdiff:= 0;
      newcmp := 0
    end;
    move(buf[nextPos], buf[currentPos + cmpdiff],
         sizeof(buf) - pred(currentPos) - cmpdiff);
    NewNextWord := nextStr;
    newflen := nextflen + cmpdiff;
    move(newflen, buf[currentPos], 2);
    buf[currentPos + 2] := newcmp;
    if nextStr > '' then begin
      nextStr := copy(nextStr, succ(newcmp), length(nextStr) - newcmp);
      move(nextStr[1], buf[currentPos + 3], length(nextStr))
    end;
    buf[currentPos + 3 + length(nextStr)] := 0
  end;
  dec(header.nWord)
end;

procedure EditableSearchBuffer.DelIndex(IndexToDel : integer);
(* Delete IndexToDel from Index *)
var BlockToDel : word;
begin
  with Indeces[IndexToDel] do begin
    BlockToDel := BlockNum;
    freemem(FirstWord, succ(length(FirstWord^)))
  end;
  move(Indeces[succ(IndexToDel)], Indeces[IndexToDel],
       (succ(header.nIndex) - IndexToDel) * sizeof(IndexTag));
  DelBlockFromSet(UsedBlock, BlockToDel);
  dec(header.nIndex);
  dec(header.nBlock)
end;

procedure EditableSearchBuffer.DelAWord(WordToDel : string);
var NextWord : string;
    ToSearch : SearchStat;
begin
  DosRequestMutexSem(SemHandle, 60000);
  if header.nIndex = 0 then exit;
  ResetSearchStat(ToSearch);
  with ToSearch do begin
    index := 0;
    SearchFirst(WordToDel, ToSearch);
    if currentStr <> WordToDel then exit;
    DelFromBlock(NextWord);
    if currentPos = 1 then
      if NextWord <> '' then ChangeIdxWord(NextWord, index)
                        else DelIndex(index)
  end;
  SaveBlock(sStat.num);
  DosReleaseMutexSem(SemHandle)
end;

procedure EditableSearchBuffer.SaveHeader;
var headertmp : LexHeader;
begin
  seek(lexfile, 0);
  blockread(lexfile, headertmp, sizeof(headertmp));
  with header do begin
    headertmp.nIndex     := nIndex;
    headertmp.nWord      := nWord;
    if version  div $100 = 0 then headertmp.nBlock := nWord
                             else headertmp.nBlock := nBlock
  end;
  seek(lexfile, 0);
  blockwrite(lexfile, headertmp, sizeof(headertmp))
end;

procedure EditableSearchBuffer.SaveIndex;
var
  idx, i : word;
  idxbuf : ^IndexBuffer;
begin
  new(idxbuf);
  fillchar(idxbuf^, sizeof(idxbuf^), 0);
  i := 1;
  for idx := 1  to header.nIndex do
    with Indeces[idx] do begin
      move(BlockNum, idxbuf^[i], 2);
      inc(i, 2);
      move(FirstWord^[1], idxbuf^[i], length(FirstWord^));
      inc(i, succ(length(FirstWord^)))
    end;
  seek(lexfile, HeaderLen);
  blockwrite(lexfile, idxbuf^, header.IndexSize);
  dispose(idxbuf)
end;

procedure EditableSearchBuffer.SaveBlock(NewBlockNum : word);
var i : longint;
begin
  with header do begin
    i := HeaderLen + IndexSize + BlockSize * NewBlockNum;
    seek(lexfile, i);
    blockwrite(lexfile, buf, BlockSize)
  end
end;

procedure SearchWords(WordToSearch : string;
                       var ToSearch : SearchStatAry;
                       var rsnum : DispItemNums);
(* Search WordToSearch  *)
var r : FoundItem;
    n, i : word;
    avail : boolean;

  procedure InsertItem;
  var i : word;
  begin
    if n = 0 then begin
      results[1] := r;
      n := 1
    end else begin
      i := n;
      while (i > 0) and (results[i].eng > r.eng) do dec(i);
      inc(i);
      if i <= n then
        move(results[i], results[succ(i)], sizeof(results[1]) * succ(n - i));
      results[i] := r;
      if n < setting.DispItemNum then inc(n)
    end
  end;

begin
  if (prewd <= WordToSearch) and
     (WordToSearch <= results[1].eng) then exit;
  Lex1.SearchFirst(WordToSearch, ToSearch[1]);
  n := 0;
  results[1].eng := '';
  results[setting.DispItemNum].eng := MaxStr^;
  r.LexNum := 1;
  Lex1.GetData(r, avail);
  while (n < setting.DispItemNum) and avail do begin
    inc(n);
    results[n] := r;
    Lex1.GetData(r, avail)
  end;
  for i := 2 to setting.lexnum do begin
    r.LexNum := i;
    with LexAdd[i]^ do begin
      SearchFirst(WordToSearch, ToSearch[i]);
      GetData(r, avail);
      while (r.eng < results[setting.DispItemNum].eng) and avail do begin
        InsertItem;
        GetData(r, avail)
      end
    end
  end;
  prewd := WordToSearch;
  rsnum := n
end;

function DoSearch(dummy : pointer) : longint;
var localHAB : HAB;
    ii       : ULONG;
    i        : integer;
    c        : char;
begin
  SearchRunning := true;
  localHAB := WinInitialize(0);
  SearchWords(wd, SearchHist, resultNum);
  if (resultNum > 0) and
    ((results[1].eng <> OldEng) or (results[1].eng = wd)) then begin
    ii := 1;
    for i := 1 to resultNum do begin
      with results[i] do begin
        if eng = wd then c := '*'
                    else c := ' ';
        MLEBuffer[ii] := c;
        inc(ii);
        move(eng[1], MLEBuffer[ii], length(eng));
        inc(ii, length(eng));
        MLEBuffer[ii] := ' ';
        inc(ii);
        move(jap[1], MLEBuffer[ii], length(jap));
        inc(ii, length(jap));
        MLEBuffer[ii] := #$0D;
        MLEBuffer[succ(ii)] := #$0A;
        inc(ii, 2);
        endpos := pred(ii)
      end
    end;
    OldEng := results[1].eng;
    WinPostMsg(hwndClient, WM_USER_Found, pred(ii), 0)
  end;
  WinTerminate(localHAB);
  SearchRunning := false;
  DosExit(EXIT_THREAD, 0)
end;

function DoTransSearch(dummy : pointer) : longint;
var localHAB : HAB;
    ii       : ULONG;
    i        : integer;
    p        : pTransFoundList;

  procedure DelFoundTrans;
  var p : pTransFoundList;
  begin
    while FoundTrans<>nil do begin
      p := FoundTrans;
      FoundTrans := p^.next;
      dispose(p)
    end
  end;

begin
  TransSearchRunning := true;
  localHAB := WinInitialize(0);
  Lex1.SearchTranslate(twd);
  for i := 2 to setting.lexnum do LexAdd[i]^.SearchTranslate(twd);
  p := FoundTrans;
  if p<>nil then begin
    ii := 1;
    while p<>nil do begin
      with p^ do begin
        move(Eng[1], MLEBufferTrans^[ii], length(Eng));
        inc(ii, length(Eng));
        MLEBufferTrans^[ii] := ' ';
        inc(ii);
        move(Trans[1], MLEBufferTrans^[ii], length(Trans));
        inc(ii, length(Trans));
        MLEBufferTrans^[ii] := #$0D;
        MLEBufferTrans^[succ(ii)] := #$0A;
        inc(ii, 2)
      end;
      p := p^.next
    end;
    WinPostMsg(hwndClient, WM_USER_TransFound, pred(ii), 0);
    DelFoundTrans
  end else
    dispose(MLEBufferTrans);
  WinTerminate(localHAB);
  TransSearchRunning := false;
  DosExit(EXIT_THREAD, 0)
end;

procedure CheckCommandLineParams;
var i, j : integer;
    dummy1 : NameStr;
    dummy2 : ExtStr;
begin
  with setting do begin
    if paramstr(1) = '/C' then i := 1 else i := 0;
    for j := succ(i) to paramcount do lexname[j - i] := paramstr(j);
    lexnum := paramcount - i;
    FSplit(paramstr(0), homedir, dummy1, dummy2)
{;   homedir := 'd:\program\lexsrch\'}
  end
end;

procedure InitConfig;
begin
  fillchar(setting, sizeof(setting), 0);
  setting.DispItemNum := 5
end;

procedure LoadConfig;
var t : text;
    i : integer;
    s : string;
begin
  with setting do begin
    assign(t, homedir + ConfigName);
{$I-}
    reset(t);
{$I+}
    if ioresult = 0 then begin
      readln(t, DispItemNum);
      readln(t, i);
      if i = 1 then RunMode := IncMode;
      readln(t, x0, y0, x1, y1);
      if lexnum = 0 then
        while not eof(t) do begin
          readln(t, s);
          if s > ' ' then begin
            inc(lexnum);
            lexname[lexnum] := s
          end
        end;
      close(t)
    end
  end
end;

function ChangeConfig(dummy : pointer) : longint;
var t : text;
    i : integer;
    sw: SWP;
    sx0, sy0, sx1, sy1 : string;
begin
  dos.ExecFlags := efSync;
  with setting do begin
    WinQueryWindowPos(hwndFrame, sw);
    with sw do begin
      str(x, sx0);
      str(y, sy0);
      str(cx, sx1);
      str(cy, sy1)
    end;
    Exec(getenv('COMSPEC'),
         '/C ' + homedir + Configurator + homedir + ConfigName +
                 ' ' + sx0 + ' ' + sy0 + ' ' + sx1 + ' ' + sy1);
    DispItemNum := 5;
    assign(t, homedir + ConfigName);
{$I-}
    reset(t);
{$I+}
    if ioresult = 0 then begin
      readln(t, DispItemNum);
      readln(t, i);
      if i = 1 then RunMode := IncMode
    end;
    close(t)
  end;
  DosExit(EXIT_THREAD, 0)
end;

procedure InitVars;
var i : integer;
    s : string;
begin
  if not (setting.lexnum in [1..MaxLex]) then DosExit(EXIT_PROCESS, 1);
  fillchar(SearchHist, sizeof(SearchHist), 0);
  results[1].eng := '';
  getmem(NullStr, 1);
  NullStr^[0] := chr(0);
  new(MaxStr);
  fillchar(MaxStr^, sizeof(MaxStr^), chr(255));
  prewd       := '';
  LexToAddNum := 0;
  with setting do begin
    Lex1.Init(lexname[1]);
    for i := 2 to lexnum do begin
      s := lexname[i];
      if s[1] <> '!' then
        new(LexAdd[i], init(s))
      else begin
        delete(s, 1, 1);
        new(LexToAdd, init(s));
        LexAdd[i]   := LexToAdd;
        LexToAddNum := i
      end
    end
  end;
  FoundTrans  := nil;
  wd     := '';
  OldEng := '';
  SearchRunning := false;
  CurrentFocusOnEntry := false;
  Pending:= 0;
  TimerID:= 0;
  MLEDataPos := 0
end;

procedure DisposeVars;
var i : integer;
begin
  for i := 2 to setting.lexnum do dispose(LexAdd[i], done);
  Lex1.Done;
  freemem(NullStr, 1);
  dispose(MaxStr)
end;

function SearchTransDlgProc(h : HWND; w : ULONG;
                            m1 : MPARAM; m2 : MPARAM) : MRESULT; CDECL;
begin
  case w of
    WM_INITDLG : begin
        WinSendDlgItemMsg(h, IDD_SearchTransEntry, EM_SETTEXTLIMIT, 255, 0);
        SearchTransDlgProc := 0
      end;
    WM_COMMAND : begin
        case LoUShort(m1) of
          DID_OK     : begin
              twd[0] := chr(WinQueryDlgItemText(h,
                                  IDD_SearchTransEntry, 255, @twd[1]));
              new(MLEBufferTrans);
              if not TransSearchRunning then
                BeginThread(nil, $10000, DoTransSearch, nil,
                                STACK_COMMITTED, TransThreadID);
              WinDismissDlg(h, 1)
            end;
          DID_Cancel : begin
              twd := '';
              WinDismissDlg(h, 0)
            end
        end;
        SearchTransDlgProc := 0
      end
  else
    SearchTransDlgProc := WinDefDlgProc(h, w, m1, m2)
  end
end;

function SearchTransResultDlgProc(h : HWND; w : ULONG;
                                  m1 : MPARAM; m2 : MPARAM) : MRESULT; CDECL;
begin
  case w of
    WM_INITDLG : begin
        WinSendDlgItemMsg(h, IDD_SearchTransResult, MLM_SETIMPORTEXPORT,
                          MPARAM(MLEBufferTrans), TransMLEBufSize);
        WinSetDlgItemText(h, IDD_SearchTransResult, '');
        WinSendDlgItemMsg(h, IDD_SearchTransResult, MLM_IMPORT,
                          MPARAM(@MLEDataPos), m2);
        SearchTransResultDlgProc := 0
      end
  else
    SearchTransResultDlgProc := WinDefDlgProc(h, w, m1, m2)
  end
end;

function PrefDlgProc(h : HWND; w : ULONG;
                     m1 : MPARAM; m2 : MPARAM) : MRESULT; CDECL;
begin
  case w of
    WM_INITDLG : begin
        WinSendDlgItemMsg(h, IDD_SetIncremental, BM_SETCHECK,
                          ord(setting.RunMode = IncMode), 0);
        PrefDlgProc := 0
      end;
    WM_COMMAND : begin
        case LoUShort(m1) of
          DID_OK     : begin
              case WinSendDlgItemMsg(h, IDD_SetIncremental,
                                     BM_QUERYCHECK, 0, 0) of
                0 : setting.RunMode := NormalMode;
                1 : setting.RunMode := IncMode
              end;
              WinDismissDlg(h, 1)
            end
        end;
        PrefDlgProc := 0
      end
  else
    PrefDlgProc := WinDefDlgProc(h, w, m1, m2)
  end
end;

function AboutDlgProc(h : HWND; w : ULONG;
                      m1 : MPARAM; m2 : MPARAM) : MRESULT; CDECL;
begin
  AboutDlgProc := WinDefDlgProc(h, w, m1, m2)
end;

function AddDlgProc(h : HWND; w : ULONG;
                    m1 : MPARAM; m2 : MPARAM) : MRESULT; CDECL;
var s, tr, ex : string;
begin
  case w of
    WM_INITDLG : begin
        WinSetWindowText(h, 'P̒ǉ');
        s := wd + chr(0);
        WinSetDlgItemText(h, IDD_EditWord, @s[1]);
        AddDlgProc := 0
      end;
    WM_COMMAND :
        case LoUShort(m1) of
          DID_OK     : begin
              s[0] := chr(WinQueryDlgItemText(h,
                                              IDD_EditWord, 255, @s[1]));
              if s > ' ' then begin
                tr[0]:= chr(WinQueryDlgItemText(h,
                                                IDD_EditResult, 255, @tr[1]));
                ex[0]:= chr(WinQueryDlgItemText(h,
                                                IDD_EditDetail, 255, @ex[1]));
                with LexToAdd^ do begin
                  AddAWord(s, tr, ex);
                  SaveHeader;
                  SaveIndex
                end;
                ResetSearchStat(SearchHist[LexToAddNum])
              end;
              WinDismissDlg(h, 1)
            end;
          DID_CANCEL  :
              WinDismissDlg(h, 0)
        end
  else
    AddDlgProc := WinDefDlgProc(h, w, m1, m2)
  end
end;

function DelDlgProc(h : HWND; w : ULONG;
                    m1 : MPARAM; m2 : MPARAM) : MRESULT; CDECL;
var s : string;
begin
  case w of
    WM_INITDLG : begin
        WinSetWindowText(h, 'P̍폜');
        s := wd + chr(0);
        WinSetDlgItemText(h, IDD_EditWord, @s[1]);
        if wd = results[1].eng then begin
          s := results[1].jap + chr(0);
          WinSetDlgItemText(h, IDD_EditResult, @s[1])
        end;
        DelDlgProc := 0
      end;
    WM_COMMAND :
        case LoUShort(m1) of
          DID_OK     : begin
              s[0] := chr(WinQueryDlgItemText(h,
                                              IDD_EditWord, 255, @s[1]));
              if s > ' ' then begin
                with LexToAdd^ do begin
                  DelAWord(s);
                  SaveHeader;
                  SaveIndex
                end;
                ResetSearchStat(SearchHist[LexToAddNum])
              end;
              WinDismissDlg(h, 1)
            end;
          DID_CANCEL  :
              WinDismissDlg(h, 0)
        end
  else
    DelDlgProc := WinDefDlgProc(h, w, m1, m2)
  end
end;

function EditDlgProc(h : HWND; w : ULONG;
                     m1 : MPARAM; m2 : MPARAM) : MRESULT; CDECL;
var s, tr, ex : string;
begin
  case w of
    WM_INITDLG : begin
        WinSetWindowText(h, 'P̏C');
        s := wd + chr(0);
        with results[1] do begin
          WinSetDlgItemText(h, IDD_EditWord, @s[1]);
          s := jap + chr(0);
          WinSetDlgItemText(h, IDD_EditResult, @s[1]);
          if extAtr > 0 then begin
            if LexNum = 1 then
              Lex1.GetExample(results[1])
            else
              LexAdd[LexNum]^.GetExample(results[1])
          end else
            Example := '';
          s := Example + chr(0);
          WinSetDlgItemText(h, IDD_EditDetail, @s[1])
        end;
        EditDlgProc := 0
      end;
    WM_COMMAND :
        case LoUShort(m1) of
          DID_OK     : begin
              s[0] := chr(WinQueryDlgItemText(h,
                                              IDD_EditWord, 255, @s[1]));
              if s > ' ' then begin
                tr[0]:= chr(WinQueryDlgItemText(h,
                                                IDD_EditResult, 255, @tr[1]));
                ex[0]:= chr(WinQueryDlgItemText(h,
                                                IDD_EditDetail, 255, @ex[1]));
                with LexToAdd^ do begin
                  DelAWord(s);
                  SaveHeader;
                  SaveIndex;
                  ResetSearchStat(SearchHist[LexToAddNum]);
                  AddAWord(s, tr, ex);
                  SaveHeader;
                  SaveIndex
                end;
                ResetSearchStat(SearchHist[LexToAddNum])
              end;
              WinDismissDlg(h, 1)
            end;
          DID_CANCEL  :
              WinDismissDlg(h, 0)
        end
  else
    EditDlgProc := WinDefDlgProc(h, w, m1, m2)
  end
end;

function DetailDlgProc(h : HWND; w : ULONG;
                       m1 : MPARAM; m2 : MPARAM) : MRESULT; CDECL;
begin
  case w of
    WM_INITDLG : begin
        WinSetDlgItemText(h, IDD_Detail, @MLEBuffer[1]);
        DetailDlgProc := 0
      end
  else
    DetailDlgProc := WinDefDlgProc(h, w, m1, m2)
  end
end;

function ClientWindowProc(h : HWND; w : ULONG;
                          m1 : MPARAM; m2 : MPARAM) : MRESULT; CDECL;
type
  pswary = array[1..2] of SWP;
var s : string;
    i : LONG;
    sw: ^pswary;
    ss: SWP;
    dx,
    dy: LONG;

  procedure MoveSizeControl(ID, dx, dy, whyd, whyd2 : LONG);
  var hh : HWND;
      ss : SWP;
  begin
    hh := WinWindowFromID(hwndFrame, ID);
    if WinQueryWindowPos(hh, ss) then
      with ss do begin
        dec(cx, whyd);     (* why?? *)
        dec(cy, whyd);     (* why?? *)
        inc(cx, dx);
        inc(x,  whyd2);    (* why?? *)
        inc(y,  dy + whyd2);  (* why?? *)
        WinSetWindowPos(hh, HWND_TOP, x, y, cx, cy, SWP_SIZE + SWP_MOVE)
      end
  end;

  procedure SizeControl(ID, dx, dy : LONG);
  var hh : HWND;
      ss : SWP;
  begin
    hh := WinWindowFromID(hwndFrame, ID);
    if WinQueryWindowPos(hh, ss) then
      with ss do begin
        inc(cx, dx);
        inc(cy, dy);
        WinSetWindowPos(hh, HWND_TOP, 0, 0, cx, cy, SWP_SIZE)
      end
  end;

  procedure MoveControl(ID, dx, dy : LONG);
  var hh : HWND;
      ss : SWP;
  begin
    hh := WinWindowFromID(hwndFrame, ID);
    if WinQueryWindowPos(hh, ss) then
      with ss do begin
        inc(x, dx);
        inc(y, dy);
        WinSetWindowPos(hh, HWND_TOP, x, y, 0, 0, SWP_MOVE)
      end
  end;

  procedure StartSearch;
  begin
    s[0] := chr(WinQueryDlgItemText(h, IDD_SearchWord, 255, @s[1]));
    if not SearchRunning then begin
      Pending := 0;
      wd      := s;
      BeginThread(nil, 8192, DoSearch, nil, CREATE_READY, ThreadID)
    end else begin
      if (Pending = 0) or (wd <> s) then Pending := 1;
      TimerID := WinStartTimer(myhab, h, ID_Timer, RetryInterval)
    end
  end;

  procedure SetDetail;
  var ii  : integer;
      s   : string;
      c   : integer;
      focus : integer;

    function GetFocus(c : integer) : integer;
    var i : integer;
    begin
      for i := resultNum downto 1 do
        if c <= results[i].endpos then GetFocus := i
    end;

  begin
    c := WinSendDlgItemMsg(h, IDD_Result, MLM_QUERYSEL,
                           MLFQS_CURSORSEL, 0);
    str(c, s);
    focus := GetFocus(c);
    ii := 1;
    with results[focus] do begin
      if extAtr > 0 then begin
        if LexNum = 1 then
          Lex1.GetExample(results[focus])
        else
          LexAdd[LexNum]^.GetExample(results[focus])
      end else
        Example := '';
      s := '[P]' + chr($0D) + chr($0A);
      move(s[1], MLEBuffer[1], length(s));
      inc(ii, length(s));
      move(eng[1], MLEBuffer[ii], length(eng));
      inc(ii, length(eng));
      MLEBuffer[ii]       := #$0D;
      MLEBuffer[succ(ii)] := #$0A;
      inc(ii, 2);
      s := '[Ӗ]' + chr($0D) + chr($0A);
      move(s[1], MLEBuffer[ii], length(s));
      inc(ii, length(s));
      move(jap[1], MLEBuffer[ii], length(jap));
      inc(ii, length(jap));
      MLEBuffer[ii]       := #$0D;
      MLEBuffer[succ(ii)] := #$0A;
      inc(ii, 2);
      s := '[]' + chr($0D) + chr($0A);
      move(s[1], MLEBuffer[ii], length(s));
      inc(ii, length(s));
      move(Example[1], MLEBuffer[ii], length(Example));
      inc(ii, length(Example));
      MLEBuffer[ii] := chr(0)
    end
  end;

  function GetClipBoardText : string;
  var pszClipTxt : PChar;
      s          : string;
  begin
    WinOpenClipbrd(myhab);
    pszClipTxt := PChar(WinQueryClipbrdData(myhab, CF_TEXT));
    if pszClipTxt <> nil then s := StrPas(pszClipTxt)
                         else s := '';
    WinCloseClipbrd(myhab);
    GetClipBoardText := s
  end;

begin
  ClientWindowProc := 0;
  case w of
    WM_CHAR    : begin
        if ((LoUShort(m1) and KC_KEYUP) = 0) and
           (HiUShort(m2) = VK_ESC) then
          WinSetDlgItemText(h, IDD_SearchWord, '');
        ClientWindowProc := WinDefDlgProc(h, w, m1, m2)
      end;
    WM_CONTROL :
        case HiUShort(m1) of
          EN_CHANGE : begin    (* only from EntryField *)
              if not CurrentFocusOnEntry then begin
                WinSetDlgItemText(h, DID_OK, '');
                CurrentFocusOnEntry := true
              end;
              if setting.RunMode = IncMode then StartSearch
            end;
          EN_SETFOCUS : begin  (* only from EntryField *)
              WinSetDlgItemText(h, DID_OK, '');
              CurrentFocusOnEntry := true
            end;
          MLN_SETFOCUS : begin (* only from MLE *)
              CurrentFocusOnEntry := false;
              WinSetDlgItemText(h, DID_OK, 'ڍ')
            end
        end;
    WM_COMMAND :
        case LoUShort(m1) of
          DID_OK     :
              if CurrentFocusOnEntry then begin
                StartSearch;
                WinSetFocus(HWND_DESKTOP,
                            WinWindowFromID(hwndFrame, IDD_SearchWord))
              end else begin
                SetDetail;
                WinDlgBox(HWND_DESKTOP, h, DetailDlgProc, NULLHANDLE,
                          IDD_DetailDlg, nil)
              end;
          IDM_SearchTrans :
                WinDlgBox(HWND_DESKTOP, h, SearchTransDlgProc, NULLHANDLE,
                          IDD_SearchTransDlg, nil);
          IDM_RunMode :
              WinDlgBox(HWND_DESKTOP, h, PrefDlgProc, NULLHANDLE,
                        IDD_RunMode, nil);
          IDM_RunSetup:
              VPBeginThread(ChangeConfig, 8192, nil);
          IDM_About   :
              WinDlgBox(HWND_DESKTOP, h, AboutDlgProc, NULLHANDLE,
                        IDD_About, nil);
          IDM_AddWord,
          IDM_DelWord,
          IDM_EditWord :
              if LexToAddNum > 0 then
                case LoUShort(M1) of
                  IDM_AddWord :
                      WinDlgBox(HWND_DESKTOP, h, AddDlgProc, NULLHANDLE,
                                IDD_Edit, nil);
                  IDM_DelWord :
                      WinDlgBox(HWND_DESKTOP, h, DelDlgProc, NULLHANDLE,
                                IDD_Edit, nil);
                  IDM_EditWord :
                      if (wd > ' ') and (wd = results[1].eng) then
                        WinDlgBox(HWND_DESKTOP, h, EditDlgProc, NULLHANDLE,
                                  IDD_Edit, nil)
                      else
                        WinMessageBox(HWND_DESKTOP, h,
                                      'ĂĈ',
                                      's', 1, MB_OK)
                end
              else
                WinMessageBox(HWND_DESKTOP, h,
                              'ύXłt@Cw肳ĂȂł',
                              '߂Ȃł', 1, MB_OK);
          IDM_SearchClipBoardText : begin
              s := GetClipBoardText;
              if s<>'' then begin
                wd := s;
                s  := s + chr(0);
                WinSetDlgItemText(h, IDD_SearchWord, @s[1]);
                WinSendMsg(h, WM_COMMAND, DID_OK, 0)
              end
            end
        end;
    WM_WINDOWPOSCHANGED : begin
        ClientWindowProc := WinDefDlgProc(h, w, m1, m2);
        sw := pointer(m1);
        with sw^[1] do
          if (fl and SWP_SIZE) <> 0 then begin
            dx := cx - sw^[2].cx;
            dy := cy - sw^[2].cy;
            if (dx <> 0) or (dy <> 0) then begin
              MoveSizeControl(IDD_SearchWord, dx, dy, 6, 3);  (* why?? *)
              SizeControl(IDD_Result, dx, dy);
              MoveControl(DID_OK, dx, dy);
              WinInvalidateRect(hwndFrame, nil, true)
            end
          end
      end;
    WM_USER_Found : begin
        WinSetDlgItemText(h, IDD_Result, '');
        WinSendDlgItemMsg(h, IDD_Result, MLM_IMPORT, MPARAM(@MLEDataPos), m1)
      end;
    WM_USER_TransFound : begin
        WinDlgBox(HWND_DESKTOP, h, SearchTransResultDlgProc, NULLHANDLE,
                  IDD_SearchTransResultDlg, pointer(m1));
        dispose(MLEBufferTrans)
      end;
    WM_INITDLG : begin
        WinSendDlgItemMsg(h, IDD_SearchWord, EM_SETTEXTLIMIT, 255, 0);
        WinSendDlgItemMsg(h, IDD_Result, MLM_SETIMPORTEXPORT,
                          MPARAM(@MLEBuffer[1]), MLEBufSize);
        hwndClient := h
      end;
    WM_TIMER :
      if Pending > 0 then
        if Pending < MaxPending then begin
          inc(Pending);
          StartSearch
        end else begin
          WinStopTimer(myhab, h, TimerID);
          TimerID := 0;
          Pending := 0;
          WinMessageBox(HWND_DESKTOP, h,
                        'Xbh~܂Ȃł()',
                        'Ȃɂ', 1, MB_OK)
        end
      else begin
        WinStopTimer(myhab, h, TimerID);
        TimerID := 0
      end;
    WM_CLOSE : begin
        if TimerID <> 0 then WinStopTimer(myhab, h, ID_Timer);
        WinPostMsg(h, WM_QUIT, 0, 0)
      end
  else
    ClientWindowProc := WinDefDlgProc(h, w, m1, m2)
  end
end;  { ClientWindowProc }


begin
  InitConfig;
  CheckCommandLineParams;
  LoadConfig;
  InitVars;

  myhab := WinInitialize(0);
  myhmq := WinCreateMsgQueue(myhab, 0);

  WinRegisterClass(myhab, ClassName, ClientWindowProc, CS_SIZEREDRAW, 0);
  hwndFrame := WinLoadDlg(HWND_DESKTOP, HWND_DESKTOP, ClientWindowProc,
                          NULLHANDLE, ControlProgramID, nil);

  fillchar(sw, sizeof(sw), 0);
  sw.hwnd := hwndFrame;
  WinAddSwitchEntry(@sw);

  hwndMenu := WinLoadMenu(hwndFrame, NULLHANDLE, ControlProgramID);

  with setting do
    if (x1 <> 0) and (y1 <> 0) then
      WinSetWindowPos(hwndFrame, HWND_TOP,
                      x0, y0, x1, y1, SWP_SIZE + SWP_MOVE);

  while WinGetMsg(myhab, myqmsg, 0, 0, 0) do
    WinDispatchMsg(myhab, myqmsg);

  WinDestroyWindow(hwndFrame);
  WinDestroyMsgQueue(myhmq);
  WinTerminate(myhab);

  DisposeVars
end.

